@@ -82,7 +82,7 @@ static const char canon_map[256] = {
    0,   0,   0,   0,   0,   0,   0,   0,    0,   0,   0,   0,   0,   0,   0,   0,
    0,   0,   0,   0,   0,   0,   0,   0,    0,   0,   0,   0,   0,  '-',  0,   0,
   '0', '1', '2', '3', '4', '5', '6', '7',  '8', '9',  0,   0,   0,   0,   0,   0,
-   0,  'a', 'b', 'c', 'd', 'e', 'f', 'g',  'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
+  '-', 'a', 'b', 'c', 'd', 'e', 'f', 'g',  'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
   'p', 'q', 'r', 's', 't', 'u', 'v', 'w',  'x', 'y', 'z',  0,   0,   0,   0,  '-',
    0,  'a', 'b', 'c', 'd', 'e', 'f', 'g',  'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
   'p', 'q', 'r', 's', 't', 'u', 'v', 'w',  'x', 'y', 'z',  0,   0,   0,   0,   0
@@ -688,7 +688,7 @@ pango_language_includes_script (PangoLanguage *language,
 static PangoLanguage **
 parse_default_languages (void)
 {
-  char *p;
+  char *p, *p_copy;
   gboolean done = FALSE;
   GArray *langs;
 
@@ -700,7 +700,7 @@ parse_default_languages (void)
   if (p == NULL)
     return NULL;
 
-  p = g_strdup (p);
+  p_copy = p = g_strdup (p);
 
   langs = g_array_new (TRUE, FALSE, sizeof (PangoLanguage *));
 
@@ -727,6 +727,8 @@ parse_default_languages (void)
 	p = end + 1;
     }
 
+  g_free (p_copy);
+
   return (PangoLanguage **) g_array_free (langs, FALSE);
 }
 
