@@ -19,7 +19,7 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include <config.h>
+#include "config.h"
 
 #include "pangofc-font.h"
 #include "pangofc-fontmap.h"
@@ -34,22 +34,8 @@
 
 enum {
   PROP_0,
-  PROP_PATTERN
-};
-
-typedef struct _GUnicharToGlyphCacheEntry GUnicharToGlyphCacheEntry;
-
-/* An entry in the fixed-size cache for the gunichar -> glyph mapping.
- * The cache is indexed by the lower N bits of the gunichar (see
- * GLYPH_CACHE_NUM_ENTRIES).  For scripts with few code points,
- * this should provide pretty much instant lookups.
- *
- * The "ch" field is zero if that cache entry is invalid.
- */
-struct _GUnicharToGlyphCacheEntry
-{
-  gunichar   ch;
-  PangoGlyph glyph;
+  PROP_PATTERN,
+  PROP_FONTMAP
 };
 
 typedef struct _PangoFcFontPrivate PangoFcFontPrivate;
@@ -57,13 +43,11 @@ typedef struct _PangoFcFontPrivate PangoFcFontPrivate;
 struct _PangoFcFontPrivate
 {
   PangoFcDecoder *decoder;
-  gpointer context_key;
-  GUnicharToGlyphCacheEntry *char_to_glyph_cache;
+  PangoFcFontKey *key;
+  PangoFcCmapCache *cmap_cache;
+  gboolean has_weak_pointer; /* have set a weak_pointer from fontmap to us */
 };
 
-#define GLYPH_CACHE_NUM_ENTRIES 256 /* should be power of two */
-#define GLYPH_CACHE_MASK (GLYPH_CACHE_NUM_ENTRIES - 1)
-
 static gboolean pango_fc_font_real_has_char  (PangoFcFont *font,
 					      gunichar     wc);
 static guint    pango_fc_font_real_get_glyph (PangoFcFont *font,
@@ -74,6 +58,10 @@ static void                  pango_fc_font_set_property (GObject          *objec
 							 guint             prop_id,
 							 const GValue     *value,
 							 GParamSpec       *pspec);
+static void                  pango_fc_font_get_property (GObject          *object,
+							 guint             prop_id,
+							 GValue           *value,
+							 GParamSpec       *pspec);
 static PangoEngineShape *    pango_fc_font_find_shaper  (PangoFont        *font,
 							 PangoLanguage    *language,
 							 guint32           ch);
@@ -103,6 +91,7 @@ pango_fc_font_class_init (PangoFcFontClass *class)
 
   object_class->finalize = pango_fc_font_finalize;
   object_class->set_property = pango_fc_font_set_property;
+  object_class->get_property = pango_fc_font_get_property;
   font_class->describe = pango_fc_font_describe;
   font_class->describe_absolute = pango_fc_font_describe_absolute;
   font_class->find_shaper = pango_fc_font_find_shaper;
@@ -114,7 +103,15 @@ pango_fc_font_class_init (PangoFcFontClass *class)
 				   g_param_spec_pointer ("pattern",
 							 "Pattern",
 							 "The fontconfig pattern for this font",
-							 G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY));
+							 G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY |
+							 G_PARAM_STATIC_STRINGS));
+  g_object_class_install_property (object_class, PROP_FONTMAP,
+				   g_param_spec_object ("fontmap",
+							"Font Map",
+							"The PangoFc font map this font is associated with (Since: 1.26)",
+							PANGO_TYPE_FC_FONT_MAP,
+							G_PARAM_READWRITE |
+							G_PARAM_STATIC_STRINGS));
 
   g_type_class_add_private (object_class, sizeof (PangoFcFontPrivate));
 }
@@ -144,7 +141,15 @@ pango_fc_font_finalize (GObject *object)
   g_slist_free (fcfont->metrics_by_lang);
 
   if (fcfont->fontmap)
-    _pango_fc_font_map_remove (PANGO_FC_FONT_MAP (fcfont->fontmap), fcfont);
+    {
+      _pango_fc_font_map_remove (PANGO_FC_FONT_MAP (fcfont->fontmap), fcfont);
+      if (priv->has_weak_pointer)
+        {
+	  priv->has_weak_pointer = FALSE;
+	  g_object_remove_weak_pointer (G_OBJECT (fcfont->fontmap), (gpointer *) (gpointer) &fcfont->fontmap);
+	}
+      fcfont->fontmap = NULL;
+    }
 
   FcPatternDestroy (fcfont->font_pattern);
   pango_font_description_free (fcfont->description);
@@ -152,7 +157,8 @@ pango_fc_font_finalize (GObject *object)
   if (priv->decoder)
     _pango_fc_font_set_decoder (fcfont, NULL);
 
-  g_free (priv->char_to_glyph_cache);
+  if (priv->cmap_cache)
+    _pango_fc_cmap_cache_unref (priv->cmap_cache);
 
   G_OBJECT_CLASS (pango_fc_font_parent_class)->finalize (object);
 }
@@ -197,11 +203,12 @@ pango_fc_font_set_property (GObject       *object,
 			    const GValue  *value,
 			    GParamSpec    *pspec)
 {
+  PangoFcFont *fcfont = PANGO_FC_FONT (object);
+
   switch (prop_id)
     {
     case PROP_PATTERN:
       {
-	PangoFcFont *fcfont = PANGO_FC_FONT (object);
 	FcPattern *pattern = g_value_get_pointer (value);
 
 	g_return_if_fail (pattern != NULL);
@@ -213,6 +220,55 @@ pango_fc_font_set_property (GObject       *object,
 	fcfont->is_hinted = pattern_is_hinted (pattern);
 	fcfont->is_transformed = pattern_is_transformed (pattern);
       }
+      goto set_decoder;
+
+    case PROP_FONTMAP:
+      {
+	PangoFcFontMap *fcfontmap = PANGO_FC_FONT_MAP (g_value_get_object (value));
+
+	g_return_if_fail (fcfont->fontmap == NULL);
+	fcfont->fontmap = (PangoFontMap *) fcfontmap;
+	if (fcfont->fontmap)
+	  {
+	    PangoFcFontPrivate *priv = fcfont->priv;
+	    priv->has_weak_pointer = TRUE;
+	    g_object_add_weak_pointer (G_OBJECT (fcfont->fontmap), (gpointer *) (gpointer) &fcfont->fontmap);
+	  }
+      }
+      goto set_decoder;
+
+    default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+      return;
+    }
+
+set_decoder:
+  /* set decoder if both pattern and fontmap are set now */
+  if (fcfont->font_pattern && fcfont->fontmap)
+    _pango_fc_font_set_decoder (fcfont,
+				pango_fc_font_map_find_decoder  ((PangoFcFontMap *) fcfont->fontmap,
+								 fcfont->font_pattern));
+}
+
+static void
+pango_fc_font_get_property (GObject       *object,
+			    guint          prop_id,
+			    GValue        *value,
+			    GParamSpec    *pspec)
+{
+  switch (prop_id)
+    {
+    case PROP_PATTERN:
+      {
+	PangoFcFont *fcfont = PANGO_FC_FONT (object);
+	g_value_set_pointer (value, fcfont->font_pattern);
+      }
+      break;
+    case PROP_FONTMAP:
+      {
+	PangoFcFont *fcfont = PANGO_FC_FONT (object);
+	g_value_set_object (value, fcfont->fontmap);
+      }
       break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -257,7 +313,7 @@ pango_fc_get_shaper_map (PangoLanguage *language)
 }
 
 static PangoEngineShape *
-pango_fc_font_find_shaper (PangoFont     *font,
+pango_fc_font_find_shaper (PangoFont     *font G_GNUC_UNUSED,
 			   PangoLanguage *language,
 			   guint32        ch)
 {
@@ -271,7 +327,7 @@ pango_fc_font_find_shaper (PangoFont     *font,
 
 static PangoCoverage *
 pango_fc_font_get_coverage (PangoFont     *font,
-			    PangoLanguage *language)
+			    PangoLanguage *language G_GNUC_UNUSED)
 {
   PangoFcFont *fcfont = (PangoFcFont *)font;
   PangoFcFontPrivate *priv = fcfont->priv;
@@ -283,6 +339,9 @@ pango_fc_font_get_coverage (PangoFont     *font,
       return _pango_fc_font_map_fc_to_coverage (charset);
     }
 
+  if (!fcfont->fontmap)
+    return pango_coverage_new ();
+
   return _pango_fc_font_map_get_coverage (PANGO_FC_FONT_MAP (fcfont->fontmap),
 					  fcfont);
 }
@@ -467,7 +526,7 @@ pango_fc_font_create_metrics_for_context (PangoFcFont   *fcfont,
   pango_layout_get_extents (layout, NULL, &extents);
 
   metrics->approximate_char_width =
-    extents.width / g_utf8_strlen (sample_str, -1);
+    extents.width / pango_utf8_strwidth (sample_str);
 
   pango_layout_set_text (layout, "0123456789", -1);
   metrics->approximate_digit_width = max_glyph_width (layout);
@@ -503,10 +562,14 @@ pango_fc_font_get_metrics (PangoFont     *font,
 
   if (!tmp_list)
     {
+      PangoFontMap *fontmap;
       PangoContext *context;
 
-      if (!fcfont->fontmap)
+      /* XXX this is racy.  because weakref's are racy... */
+      fontmap = fcfont->fontmap;
+      if (!fontmap)
 	return pango_font_metrics_new ();
+      fontmap = g_object_ref (fontmap);
 
       info = g_slice_new0 (PangoFcMetricsInfo);
 
@@ -515,12 +578,13 @@ pango_fc_font_get_metrics (PangoFont     *font,
 
       info->sample_str = sample_str;
 
-      context = pango_fc_font_map_create_context (PANGO_FC_FONT_MAP (fcfont->fontmap));
+      context = pango_font_map_create_context (fontmap);
       pango_context_set_language (context, language);
 
       info->metrics = pango_fc_font_create_metrics_for_context (fcfont, context);
 
       g_object_unref (context);
+      g_object_unref (fontmap);
     }
 
   return pango_font_metrics_ref (info->metrics);
@@ -551,23 +615,25 @@ static guint
 pango_fc_font_real_get_glyph (PangoFcFont *font,
 			      gunichar     wc)
 {
+  PangoFcFontPrivate *priv = font->priv;
   FT_Face face;
   FT_UInt index;
 
   guint idx;
-  GUnicharToGlyphCacheEntry *entry;
+  PangoFcCmapCacheEntry *entry;
 
-  PangoFcFontPrivate *priv = font->priv;
 
-  if (G_UNLIKELY (priv->char_to_glyph_cache == NULL))
+  if (G_UNLIKELY (priv->cmap_cache == NULL))
     {
-      priv->char_to_glyph_cache = g_new0 (GUnicharToGlyphCacheEntry, GLYPH_CACHE_NUM_ENTRIES);
-      /* Make sure all cache entries are invalid initially */
-      priv->char_to_glyph_cache[0].ch = 1; /* char 1 cannot happen in bucket 0 */
+      priv->cmap_cache = _pango_fc_font_map_get_cmap_cache ((PangoFcFontMap *) font->fontmap,
+							    font);
+
+      if (G_UNLIKELY (!priv->cmap_cache))
+	return 0;
     }
 
-  idx = wc & GLYPH_CACHE_MASK;
-  entry = priv->char_to_glyph_cache + idx;
+  idx = wc & CMAP_CACHE_MASK;
+  entry = priv->cmap_cache->entries + idx;
 
   if (entry->ch != wc)
     {
@@ -741,6 +807,9 @@ pango_fc_font_kern_glyphs (PangoFcFont      *font,
   FT_Vector kerning;
   int i;
   gboolean hinting = font->is_hinted;
+  gboolean scale = FALSE;
+  double xscale = 1;
+  PangoFcFontKey *key;
 
   g_return_if_fail (PANGO_IS_FC_FONT (font));
   g_return_if_fail (glyphs != NULL);
@@ -755,6 +824,20 @@ pango_fc_font_kern_glyphs (PangoFcFont      *font,
       return;
     }
 
+  /* This is a kludge, and dupped in pango_ot_buffer_output().
+   * Should move the scale factor to PangoFcFont layer. */
+  key = _pango_fc_font_get_font_key (font);
+  if (key) {
+    const PangoMatrix *matrix = pango_fc_font_key_get_matrix (key);
+    PangoMatrix identity = PANGO_MATRIX_INIT;
+    if (G_UNLIKELY (matrix && 0 != memcmp (&identity, matrix, 2 * sizeof (double))))
+      {
+	scale = TRUE;
+	pango_matrix_get_font_scale_factors (matrix, &xscale, NULL);
+	if (xscale) xscale = 1 / xscale;
+      }
+  }
+
   for (i = 1; i < glyphs->num_glyphs; ++i)
     {
       error = FT_Get_Kerning (face,
@@ -768,6 +851,8 @@ pango_fc_font_kern_glyphs (PangoFcFont      *font,
 
 	if (hinting)
 	  adjustment = PANGO_UNITS_ROUND (adjustment);
+	if (G_UNLIKELY (scale))
+	  adjustment *= xscale;
 
 	glyphs->glyphs[i-1].geometry.width += adjustment;
       }
@@ -821,21 +906,21 @@ _pango_fc_font_set_decoder (PangoFcFont    *font,
     g_object_ref (priv->decoder);
 }
 
-gpointer
-_pango_fc_font_get_context_key (PangoFcFont *fcfont)
+PangoFcFontKey *
+_pango_fc_font_get_font_key (PangoFcFont *fcfont)
 {
   PangoFcFontPrivate *priv = fcfont->priv;
 
-  return priv->context_key;
+  return priv->key;
 }
 
 void
-_pango_fc_font_set_context_key (PangoFcFont *fcfont,
-				gpointer     context_key)
+_pango_fc_font_set_font_key (PangoFcFont    *fcfont,
+			     PangoFcFontKey *key)
 {
   PangoFcFontPrivate *priv = fcfont->priv;
 
-  priv->context_key = context_key;
+  priv->key = key;
 }
 
 static FT_Glyph_Metrics *
